/*          l

    PPS marker byte:
      0     = PPS pulse is not valid
      1..31 = ethernet frame ends n*500 usec after the pps pulse (500us .. 15.5 ms)

      

    sw
     1    PPS polarity
     2    enable PPS frames (PPS)
     3    enable AUX frames (PPS, PPS timestamps, RS232 input data)
     4    500us
     5    1ms
     6    2ms
     7






*/

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#include "board.h"

#include "eep.h"


#define eeprom_magic_cookie 12349


void setdac(unsigned v);

unsigned dacsetpoint;

volatile unsigned rxbuf[256];
volatile unsigned char rxbufi,rxbufo;
ISR(SIG_UART0_RECV)
{
  unsigned char p=rxbufi;
  rxbuf[p]=UDR;
  ++p;
  rxbufi=p;
}

unsigned char rxready(void)
{
  return rxbufi-rxbufo;
}

unsigned char getc(void)
{
  unsigned char d;
  unsigned char p=rxbufo;
  d=rxbuf[p];
  ++p;
  rxbufo=p;
  return d;
}





static inline unsigned long BEL(unsigned long i)
{
  union {unsigned long i; unsigned char b[4];} a,b;
  a.i=i;
  b.b[0]=a.b[3];
  b.b[1]=a.b[2];
  b.b[2]=a.b[1];
  b.b[3]=a.b[0];
  return b.i;
}


// send functions take on character payload and returns the completion timestamp
unsigned sendpk(unsigned char d); //tager 90 us = 11kbyte/sec max
unsigned sendpps(unsigned char d); //tager 90 us = 11kbyte/sec max
unsigned sendpk_precise(unsigned char d);
unsigned sendpps_precise(unsigned char d);
void ether_sync(void);



void sendpps_burststart(void); //pointer only!
void sendpps_burstend(void); //pointer only!
void sendpk_burststart(void); //pointer only!
void sendpk_burstend(void); //pointer only!

volatile unsigned counterhigh;
volatile unsigned char nlptimer;
volatile unsigned char blanker;
volatile unsigned long eeupdatetimer;
#ifdef ENABLE_AUTO_EDGE
volatile unsigned char levelautodetectCT;
#endif

#ifdef OCXOTYPE
volatile unsigned char irq_updates_dac;
volatile unsigned char irq_dac_dither;
unsigned char irq_dac_accum;
#endif


ISR(TIMER1_OVF_vect) //305 Hz
{
  unsigned char d;
  ++counterhigh;
  ++nlptimer;
  d=blanker;
  if (d) blanker=--d;
  sei();
#ifdef ENABLE_AUTO_EDGE
  //PD6 input = ICP1
  //ramp autodetect up and down to
  d=levelautodetectCT;
#ifdef BOARD168
  if (PINB&(1<<0)) {
#else
  if (PIND&(1<<6)) {
#endif
    ++d;
    if (d) levelautodetectCT=d;
  } else {
    if (d) levelautodetectCT=--d;
  }
#endif
  ++eeupdatetimer;
#ifdef OCXOTYPE
  //dithering to get higher DAC resolution
  if (irq_updates_dac) setdac(dacsetpoint);
#endif
}


unsigned long getclocktime(void) //only non-irq usage
{
  union { unsigned long t; unsigned u[2];} t;
  unsigned char TIFRcopy;
  cli();
  t.u[0]=TCNT1;
  t.u[1]=counterhigh; //burns 4 clocks, TIFR has updated
  TIFRcopy=TIFR1;
  sei();
  if (!(t.u[0]&0x8000)) //timer is low
    if (TIFRcopy&(1<<TOV1)) //overflow is pending
      ++t.u[1];
  return t.t;
}


volatile unsigned CapEdge;

//used by transmitter functions
volatile unsigned PreciseTxDelay;
volatile unsigned PreciseTxTime;

volatile unsigned LastTxCompletionDelay;

volatile union { unsigned long ul; unsigned u[2]; unsigned char b[4];} lastts;
volatile unsigned char lastts_valid;
volatile unsigned char lastts_valid_for_loop;

volatile unsigned char pps_transmit_data;
volatile unsigned char pps_transmit_slot;
volatile unsigned char pps_on_aux;

//capture1 has higher priority than overflow1
ISR(TIMER1_CAPT_vect) //falling edge
{
  unsigned edge;
  unsigned left;
  if (blanker) return;
  TIMSK1&=~(1<<TICIE1); //edge irq
  blanker=100; // 305Hz => 327 ms

  //store a 32 bit timestamp for (optionally fake) VCXO operation
  edge=ICR1;
  left=counterhigh;
  if (!(edge&0x8000)) //timer is low
    if (TIFR1&(1<<TOV1)) //overflow is pending
      ++left;
  lastts.u[0]=edge;
  lastts.u[1]=left;
  lastts_valid=1;
  lastts_valid_for_loop=1;

  //calculate the time at which to run the transmitter
  edge+=PreciseTxDelay;
  PreciseTxTime=edge;

  //normal processing until we get close. The margin must be larger than
  //the longest interrupt
  do {
    asm volatile("sei\r\nnop\r\nnop\r\nnop\r\ncli\r\n"); //take pending interrupts
    left=edge-TCNT1;
    if (left>10000) { //more than 500 usec to frame transmit, it is okay to send NLPs.
      if (nlptimer>=5) {
        nlptimer=0;
        ether_sync();
      }
    }
  } while (left>1000); //transmit is less than 50 us away. no more interrupts.
  if (ENABLE_PPS_ON_AUX) {
    left = sendpps_precise(pps_transmit_data); //does cycle accurate transmit, returns EOF stamp
    LastTxCompletionDelay = left-lastts.u[0]; //ICP(filtered)-eof time
  }
  if (pps_on_aux) {
    //aux pps is located 200us after the real pps
    unsigned addv = 4000
     + ((unsigned char*)sendpps_burstend-(unsigned char*)sendpps_burststart)
     - ((unsigned char*)sendpk_burstend-(unsigned char*)sendpk_burststart);
    PreciseTxTime+=addv;
    sendpk_precise(0xBE);
    PreciseTxTime-=addv;
  }
  TIMSK1|=(1<<TICIE1); //enable edge irq
}


void backtask(void)
{
    cli(); //305Hz count = 5 cts pr sync
    if (nlptimer>=5) {
      nlptimer=0;
      ether_sync();
    }
    sei();
}


void telebyte(unsigned char d)
{
  unsigned t,dt;
  cli();
  t=TCNT1;
  if (nlptimer>=5) {
    nlptimer=0;
    ether_sync();
  }
  asm volatile("sei\r\nnop\r\nnop\r\nnop\r\nnop\r\ncli\r\n"); //take interrupts
  sendpk(d);
  sei();

  do {
    cli();
    dt=TCNT1;
    sei();
    dt-=t;
  } while (dt<=20000); //1 ms transmit interval as the fastest permissible packet rate.
}

void debugdata(unsigned char *d, unsigned char n)
{
  telebyte(0xBD);
  telebyte(0x1F);
  while (n) {
    telebyte(*d++);
    --n;
  }
  telebyte(0);
}

void debugtext(char *memap)
{
  unsigned char n;
  unsigned p;
  telebyte(0xBD);
  telebyte(0x1F);

  p=(unsigned)memap;
  do telebyte((n=__LPM((p++)))); while (n);
}


void debugnumberUL(unsigned long ival)
{
  unsigned char buf[10];
  unsigned char n;
  n=sizeof buf;
  while (n) {
    --n;
    buf[n]=48+(ival%10);
    ival/=10;
  }
  while ((n<((sizeof buf)-1)) && (buf[n]==48)) ++n; //skip leading zeros
  debugdata(buf+n,(sizeof buf)-n);
}

void debugnumberSL(signed long ival)
{
  if (ival<0) {
    debugtext(PSTR("-"));
    ival=-ival;
  }
  debugnumberUL(ival);
}

void debugnumberU(unsigned ival)
{
  debugnumberUL(ival);
}

void debugnumberS(signed ival)
{
  debugnumberSL(ival);
}

#define COMPUTE_BARRIER(x) __asm__ volatile (""::"r"(x))

#ifdef OCXOTYPE
unsigned long daccount_pr_50ns_times_64;

void setdac(unsigned v)
{
  unsigned char ca,cb,m;
  //divider is 2M2 and 10k. That's a 221 divider.
  ca=(v%221)+1;
  cb=v/221;

  //dither code.
  if (irq_updates_dac) {
    m=irq_dac_dither;
    m+=irq_dac_accum;
    if (m&64) ++ca;
    irq_dac_accum=m&63;
  }

  COMPUTE_BARRIER(ca);
  COMPUTE_BARRIER(cb);
  cli();
  while (TCNT0>240) {
    sei();
    asm volatile("nop\r\nnop\r\nnop\r\n"::);
    cli();
  }
  OCR0A=ca;
  OCR0B=cb;
  sei();
}
#endif

signed meas_ofs(unsigned char sek) //for initial calibration
{
  unsigned udt;
  signed dt;

  lastts_valid=0;
  do { backtask(); } while (!lastts_valid);
  udt=lastts.u[0];
  do {
    lastts_valid=0;
    do { backtask(); } while (!lastts_valid);
    //20M cts pr puls = 1312D00
    udt+=0x2D00;
  } while (--sek);
  lastts_valid=0;
  udt=lastts.u[0]-udt;
  //udt er positiv hvis frekvens er for hj
  dt=udt;
  return dt;
}

#ifdef OCXOTYPE
void frq_adjust(signed ofs) //for initial calibration
{
  unsigned step=1024;
  unsigned char act,tims;
  unsigned char last=2;
  signed m;
  tims=1;
  do {
    m=meas_ofs(tims);
    if (m>(ofs*tims)) act=1; else act=0;
    if (act) dacsetpoint-=step; else dacsetpoint+=step;
    if (dacsetpoint>56355) {dacsetpoint=56355; return;} //oops.
    if (dacsetpoint>60000) {dacsetpoint=0; return;} //oops.
    setdac(dacsetpoint);
    if (act!=last) {
      step>>=1;
      last=act;
      if (tims<8) tims<<=1;
    }
  } while (step>16);
}
#endif

#define EA_cookie 32
#define EA_dacsetp 34
#define EA_dp50nsm64 36
#define EA_storea 40
#define EA_storeb 44

#ifdef OCXOTYPE
void cal_dacgain(void) //perform initial calibration
{
  unsigned ddac;
  unsigned df;
  unsigned k;
  signed mof;

  ee_read(&k, EA_cookie, sizeof k);
  if (k==eeprom_magic_cookie) {
    debugtext(PSTR("cal_dacgain(): parameters read from EEPROM\n"));
    ee_read(&daccount_pr_50ns_times_64, EA_dp50nsm64, sizeof daccount_pr_50ns_times_64);
    ee_read(&dacsetpoint, EA_dacsetp, sizeof dacsetpoint);
    setdac(dacsetpoint);
    return;
  }
  debugtext(PSTR("cal_dacgain(): Not calibrated.\n"));

  debugtext(PSTR(" Calibrating 2ppm high.."));
  frq_adjust(40); //2 ppm high
  debugtext(PSTR("done\n  DAC value: "));
  debugnumberU(dacsetpoint);
  debugtext(PSTR("\n  Frequency offset (320/ppm): "));
  mof=meas_ofs(16);
  debugnumberS(mof);
  df=mof;
  debugtext(PSTR("\n Calibrating 2ppm low.."));
  ddac=dacsetpoint;

  frq_adjust(-40); //2 ppm low
  debugtext(PSTR("done\n  DAC value: "));
  debugnumberU(dacsetpoint);
  debugtext(PSTR("\n Frequency Offset (320/ppm): "));
  mof=meas_ofs(16);
  debugnumberS(mof);
  ddac-=dacsetpoint;
  df-=mof;
  //har ddac og df. df er ca. 80.
  daccount_pr_50ns_times_64 = (1024UL*ddac)/df; //16 sek*64

  debugtext(PSTR("\n Calibrating for 0ppm.."));
  ee_write(EA_dp50nsm64,&daccount_pr_50ns_times_64, sizeof daccount_pr_50ns_times_64);
  frq_adjust(0); //initial value
  ee_write(EA_dacsetp,&dacsetpoint, sizeof dacsetpoint);

  k=eeprom_magic_cookie;
  ee_write(EA_cookie,&k, sizeof k);
  debugtext(PSTR("done\n  DAC value: "));
  debugnumberU(dacsetpoint);
  debugtext(PSTR("\ncal_dacgain(): done.\n"));

  //smash the runtime updated trim
  k=0;
  ee_write(EA_storea,&k, sizeof k);
  k=0x5555;
  ee_write(EA_storeb,&k, sizeof k);
}
#endif


/*
   line fit:

       n*sum(xi*yi) - sum(xi)*sum(yi)
   b = ------------------------------
        n*sum(xi^2) - sum(xi)^2

   x numbers: -30..0..+30 (61 seconds)
*/

struct {
  signed n;
  signed xi,yi;
  signed long X;  //sum(xi)
  signed long XX; //sum(xi^2)
  signed long Y;  //sum(yi)
  signed long XY; //sum(xi*yi)
} linefit;

void fit_clear(void)
{
  linefit.n=0;
  linefit.xi=-30;
  linefit.yi=0;
  linefit.X=0;
  linefit.XX=0;
  linefit.Y=0;
  linefit.XY=0;
}

void fit_add(signed delta)
{
  ++linefit.n;
  linefit.yi+=delta;
  linefit.X+=linefit.xi;
  linefit.XX+=((signed long)linefit.xi)*linefit.xi;
  linefit.Y+=linefit.yi;
  linefit.XY+=((signed long)linefit.xi)*linefit.yi;
  ++linefit.xi;
}

signed long fit_calc(void)
{
/*     n*sum(xi*yi) - sum(xi)*sum(yi)
   b = ------------------------------
        n*sum(xi^2) - sum(xi)^2

   theoretical resolution = 50ns/60s/sqrt(60)=  1.08E-10

             */

  signed long top,bot;

  top=linefit.n*linefit.XY - ((long)linefit.X)*((long)linefit.Y);
  bot=linefit.n*((long)linefit.XX) - ((long)linefit.X)*((long)linefit.X);
  //top/bot will be in 50ns steps per second. Which is poor resolution.
  //we need ppb's.
  //at -30..+30 the denominator becomes 18910.
  //50ppb in top is also 18910.

  // if we want to multiply by 10, that is a multiplier of 500.
  //maximum top value = 4294967.294
  while ((top>4294967)||(top<-4294967)) {
    top/=2;
    bot/=2;
  }
  top*=500;
  return top/bot; //ppb offset *10.
}







unsigned long previous_pps_stamp/*, last_pps_length*/;
#ifdef OCXOTYPE
signed long dac_control;
signed long dac_integrator;
#endif

int main(void)
{
  // initsystimer();
  //pnetinit();

  CONFIG_PWR_LED();

  TCCR1A=0;
  TCCR1B=0x81; //ck1 falling edge trigger w. noisecanceler

  TIMSK1|=(1<<TOIE1); //enable overflow count
  TIMSK1|=(1<<TICIE1); //enable edge irq

  //OC2B refout on OC2B
#ifdef OCXOTYPE
 #ifdef BOARD168
  TCCR2A=(1<<COM2B0)|(1<<WGM21);  //toggle on match
  TCCR2B=(1<<CS20); //20MHz clock
  OCR2A=0;
  OCR2B=0;
  DDRD|=(1<<3);

  TCCR0A=(1<<WGM00)|(1<<WGM01) | (1<<COM0A1) | (1<<COM0B1); //clear on match pwm mode
  TCCR0B=(1<<CS00);
  dacsetpoint=20000; //effectively midrange - the varactor is non-linear.
  setdac(20000);
  DDRD|=(1<<5);
  DDRD|=(1<<6);
 #endif
#endif



#ifdef ENABLE_AUTO_EDGE
  levelautodetectCT=127;
#endif

  //min. delay 1 packet tx time plus margin
  //100 us packet = 2000 clocks. 4000 clocks = 200us

  //6 less for the synchronizer delay


  pps_transmit_slot=PPS_TRANSMIT_SLOT; //1..15 = 0.5 ..15.5 ms
  pps_transmit_data=pps_transmit_slot;


  //this one is fishy! here gcc apparently does subtraction on word addresses.
  //do check with a scope when upgrading the compiler! works with 20090306RC.

  PreciseTxDelay = (10000UL*pps_transmit_slot)
                   - 4
                   - ((unsigned char*)sendpps_burstend-(unsigned char*)sendpps_burststart)-148;
  sei();

  while (counterhigh<610) backtask(); //2 sec power up to let switch get link.

  debugtext(PSTR("EtherPPS 0.01 (C) 2009 Kasper Kjeld Pedersen\n"));

#ifdef OCXOTYPE
  cal_dacgain();
  //load previous control loop value.
  //dac_integrator must be same type as dac_control. used here as scratch
  //space.
  ee_read(&dac_control, EA_storea, sizeof dac_control);
  ee_read(&dac_integrator, EA_storeb, sizeof dac_integrator);
  if (dac_control!=dac_integrator) {
    //stored value is invalid. use calibration setpoint
    dac_control = 64UL * dacsetpoint;
    debugtext(PSTR("main(): calculated dac_control\n"));
  }
  dac_integrator=0;

  debugtext(PSTR("main(): Startup parameters are\n DAC counts per ppm trim: "));
  //dac gain is stored as 64 times the dac counts required to tune 1/20 ppm.
  //that's 3.2 times the counts required to tune 1ppm
  debugnumberSL((daccount_pr_50ns_times_64*10)/32);

  debugtext(PSTR("\n Initial DAC value: "));
  //dac gain is stored as 64 times the dac counts required to tune 1/20 ppm.
  //that's 3.2 times the counts required to tune 1ppm
  debugnumberSL(dacsetpoint);
  debugtext(PSTR("\nmain(): loop active.\n"));
#else
  debugtext(PSTR("\nmain(): plain pps mode\n"));
#endif

  pps_on_aux=ENABLE_PPS_ON_AUX;
  fit_clear();
  eeupdatetimer=0;
  for(;;) {
    asm volatile ("nop");
    asm volatile ("nop");
#ifdef ENABLE_AUTO_EDGE
    //auto edge selection based on duty cycle
    if (levelautodetectCT==0) if (!(TCCR1B&(1<<ICES1))) {
      //mostly low level and we are not rising edge sensitive
      TIMSK1&=~(1<<TICIE1); //disable edge irq
      TCCR1B|=(1<<ICES1);  //change polarity
      asm volatile("nop\r\nnop\r\nnop\r\nnop\r\nnop\r\nnop\r\n"); //for the filter
      TIFR1=(1<<ICF1);      //clear irq if any
      TIMSK1|=(1<<TICIE1);  //enable edge irq
    }
    if (levelautodetectCT==255) if (TCCR1B&(1<<ICES1)) {
      //mostly high level and we are not falling edge sensitive
      TIMSK1&=~(1<<TICIE1); //disable edge irq
      TCCR1B&=~(1<<ICES1);  //change polarity
      asm volatile("nop\r\nnop\r\nnop\r\nnop\r\nnop\r\nnop\r\n"); //for the filter
      TIFR1=(1<<ICF1);      //clear irq if any
      TIMSK1|=(1<<TICIE1);  //enable edge irq
    }
#endif
    backtask(); //ethernet sync among other things

    if (lastts_valid_for_loop) {
      signed long dt;
#ifdef OCXOTYPE
      signed long tm;
      unsigned char dithernew;
      unsigned dacnew;
#endif      

      lastts_valid_for_loop=0;
      dt=lastts.ul - previous_pps_stamp;
      previous_pps_stamp=lastts.ul;
      dt-=20000000;
      if ((dt>-1000)&&(dt<1000)) { //50 usec window, pps is good.
        fit_add(dt);
#ifdef OCXOTYPE
        dt*=daccount_pr_50ns_times_64;
        dt/=100; //time constant: 100 seconds.
        //positive dt means we are fast.
        dac_control -= dt;
        if (dac_control<0) dac_control=0;
        tm=56355UL*64UL;
        if (dac_control>tm) dac_control=tm;
        dacnew=dac_control>>6;
        dithernew=((unsigned char)dac_control)&63;
        COMPUTE_BARRIER(dacnew);
        COMPUTE_BARRIER(dithernew);
        cli();
        irq_dac_dither=dithernew;
        *((volatile unsigned*)&dacsetpoint)=dacnew;
        sei();
        irq_updates_dac=1;
        //setdac(dacsetpoint); now the IRQ handles this.
        //305Hz, update every 60 minutes with good PPS
        if (eeupdatetimer>1098000) {
          //100k writes*1hr = 100000hrs = 11.4 years.
          eeupdatetimer=0;
          ee_write(EA_storea,&dac_control, sizeof dac_control);
          ee_write(EA_storeb,&dac_control, sizeof dac_control);
        }
#endif
      } else {
        fit_clear();
        fit_add(0);
//        eeupdatetimer=0; //uncomment to require 1 hour of continous pps to update eeprom
      }
    }


    //if we have telemetry in the buffer, send a telemetry frame
    if (lastts_valid) {
      if (!blanker) {
        lastts_valid=0;
        //send PPS timestamp (32 bit) relative to local timebase.
        //allows the recipient to do jitter analysis
	  	telebyte(0xBD);
      	telebyte(20); //code 1: pps interval data, aux pps code data
      	telebyte(lastts.b[3]);
        telebyte(lastts.b[2]);
        telebyte(lastts.b[1]);
        telebyte(lastts.b[0]);
        telebyte(pps_transmit_data);

//      setdac(step);
//      step+=500;
//      if (step>=56355) step=0;
        if (linefit.n>60) { // -30..0..+30
          union { signed long sl; unsigned u[2]; unsigned char b[4];} Q;
          union { unsigned u; unsigned char b[2];} R;
          Q.sl=fit_calc();
	  	  telebyte(0xBD);
      	  telebyte(21); //code 2: frequency error and DAC information
          telebyte(Q.b[3]);
          telebyte(Q.b[2]);
          telebyte(Q.b[1]);
          telebyte(Q.b[0]);
          R.u=dacsetpoint;
          telebyte(R.b[1]);
          telebyte(R.b[0]);
          fit_clear();
        }
      }
    }

    //uart data
    if (rxready()) {
      if (!blanker) {
        unsigned char d=getc();
        if (d==0xBD) {    //escape character
          telebyte(0xBD);
          telebyte(0);
        }
        else if (d==0xBE) { //aux pps mark reserved character
          telebyte(0xBD);
          telebyte(1);
        }
        else telebyte(d);
      }
    }

    if (!blanker) POWER_LED_ON();
    else POWER_LED_OFF();

  }

}

